from LovelacePM import *
import numpy as np
from math import tan, radians

if multiprocess_guard():
    ordir=os.getcwd()
    os.chdir(os.path.dirname(os.path.abspath(__file__)))

    b=1.1963; croot=0.806; taper=0.56; sweep=26.7
    alpha=5.0; Uinf=10.0

    root_sect=wing_section(afl='onerad', c=croot, xdisc=30, sweep=sweep) #CA_position set to origin as default
    left_tip_sect=wing_section(afl='onerad', c=croot*taper, CA_position=np.array([b*tan(radians(sweep)), -b, 0.0]), closed=True, xdisc=30, sweep=sweep)
    right_tip_sect=wing_section(afl='onerad', c=croot*taper, CA_position=np.array([b*tan(radians(sweep)), b, 0.0]), closed=True, xdisc=30, sweep=sweep)

    sld=Solid()

    left_wingquad=wing_quadrant(sld, sect1=left_tip_sect, sect2=root_sect)
    right_wingquad=wing_quadrant(sld, sect1=root_sect, sect2=right_tip_sect)
    wng=wing(sld, wingquads=[left_wingquad, right_wingquad])

    acft=aircraft(sld, elems=[wng])
    acft.edit_parameters({'a':alpha, 'Uinf':Uinf, 'M':0.7})
    acft.plot_input()

    wng.patchcompose(ydisc=50)
    acft.addwake()
    acft.plotgeometry()

    acft.eulersolve()
    acft.forces_report()
    acft.stabreport()
    plot_Cls(sld, wings=[wng])
    plot_Cds(sld, wings=[wng])
    plot_Cms(sld, wings=[wng])
    plot_Cps(sld, elems=[wng])

    results=acft.design_derivatives(wings=[wng])
    print(results)

    os.chdir(ordir)


import numpy as np
import numpy.linalg as lg
import scipy.linalg as slg
import scipy.linalg.lapack as slapack
import scipy.sparse.linalg as splg
from math import *
import matplotlib.pyplot as plt
from mpl_toolkits import mplot3d
import scipy.sparse as sps
import scipy.interpolate as sinterp
import time as tm
import os
import dill #definitions for lambda pickling
import cloudpickle as pck
import multiprocessing as mp

import pyfdyn

from paneller import *
from body import *
from wing import *

class aircraft: #class to ease certain case studies for full aircraft
    #generate aircraft BEFORE patchcompose functions
    def parameter_report(self):
        if self.sld.runme:
            print('Freestream parameters:')
            print('%10s %10s %10s %10s %10s %10s %10s' % ('a', 'b', 'p', 'q', 'r', 'Uinf', 'M'))
            print('%10f %10f %10f %10f %10f %10f %10f' % (degrees(self.a), degrees(self.b), self.p, self.q, self.r, self.Uinf, self.M))
            print('Environment parameters:')
            print('%10s %10s %10s %10s' % ('rho', 'mu', 'g', 'gamma'))
            print('%10f %10f %10f %10f' % (self.rho, self.mu, self.g, self.gamma))
            print('Geometry parameters:')
            print('%10s %10s %10s %10s' % ('Sref', 'cref', 'bref', 'AR'))
            print('%10f %10f %10f %10f' % (self.Sref, self.cref, self.bref, self.AR))
            if self.hascontrol():
                print('Control parameters:')
                controlnames_str=''
                controlvals_str=''
                for cname in self.controlset:
                    controlnames_str+='%10s' % cname
                    controlvals_str+='%10f' % degrees(self.controlset[cname].state)
                print(controlnames_str)
                print(controlvals_str)
            if not self.massavailable:
                print('Mass parameters not yet present')
            else:
                print('Mass parameters:')
                print('CG=(%10f %10f %10f)' % (self.CG[0], self.CG[1], self.CG[2]))
                print('%10s %10s %10s %10s %10s %10s %10s' % ('m', 'Ixx', 'Iyy', 'Izz', 'Ixy', 'Ixz', 'Iyz'))
                print('%10f %10f %10f %10f %10f %10f %10f' % (self.m, self.Inertia[0, 0], self.Inertia[1, 1], self.Inertia[2, 2], \
                    self.Inertia[0, 1], self.Inertia[0, 2], self.Inertia[1, 2]))
    def addmass(self, m=0.0, Ixx=0.0, Iyy=0.0, Izz=0.0, Ixy=0.0, Ixz=0.0, Iyz=0.0, echo=True): #sets mass variables for the aircraft
        if self.sld.runme:
            self.m=m
            self.Inertia=np.array([[Ixx, Ixy, Ixz], [Ixy, Iyy, Iyz], [Ixz, Iyz, Izz]])
            self.massavailable=True
            if echo:
                self.parameter_report()
    def dynamic_simulation(self, nstep=2e5, dt=1e-4, start_point=np.array([0.0, 0.0, 0.0]), \
        perturbations={}, trim=True, visc=True, onboard=np.zeros(3)):
        if self.sld.runme:
            if not self.massavailable:
                print('WARNING: dynamic analysis requested without any mass data. Returning empty arrays')
                return np.array([]), np.array([]), np.array([]), np.array([])
            keys=['phi', 'theta', 'psi', 'u', 'v', 'w', 'p', 'q', 'r']
            for par in keys:
                if not par in perturbations:
                    perturbations[par]=0.0
            perturbations=np.array([start_point[0], start_point[1], start_point[2], \
                perturbations['phi'], perturbations['theta'], perturbations['psi'], \
                    perturbations['u'], perturbations['v'], perturbations['w'], \
                        perturbations['p'], perturbations['q'], perturbations['r']])
            coeffs=np.zeros(6)
            derivs=np.zeros((5, 6))
            if not trim:
                coeffs=np.array([-self.CX, self.CY, -self.CZ, -self.Cl, self.Cm, -self.Cn]) #note that we're using stability coordinates here
                if visc:
                    coeffs+=np.array([-self.dCX, self.dCY, -self.dCZ, -self.dCl, self.dCm, -self.dCn])
            i=0
            for par in ['a', 'b', 'p', 'q', 'r']:
                derivs[i, :]=np.array([-self.stabderivative_dict[par]['dCX'], self.stabderivative_dict[par]['dCY'], -self.stabderivative_dict[par]['dCZ'], \
                    -self.stabderivative_dict[par]['dCl'], self.stabderivative_dict[par]['dCm'], -self.stabderivative_dict[par]['dCn']])
                i+=1
            external_history, alpha_history, beta_history, euler_history, time_history=pyfdyn.tstep_solve(int(nstep), dt, self.rho, self.Uinf, self.g, \
                self.Sref, self.cref, self.bref, perturbations, onboard, self.Inertia, self.m, \
                    coeffs, derivs)
            return external_history, alpha_history, beta_history, euler_history, time_history
        else:
            return np.array([]), np.array([]), np.array([]), np.array([]), np.array([])
    def balance(self, SM=0.1, echo=True): #balancing the aircraft for given static margin
        if self.sld.runme:
            if not self.stabavailable:
                print('WARNING: stability derivatives not available for balancing. Performing in place')
                self.calcstab(echo=echo)
            loc_SM=-self.stabderivative_dict['a']['dCm']/self.stabderivative_dict['a']['dCL']
            self.transp_byvec(-(SM-loc_SM)*np.array([1.0, 0.0, 0.0])*self.cref)
            new_SM=-self.stabderivative_dict['a']['dCm']/self.stabderivative_dict['a']['dCL']
            if echo:
                print('==========Balancing========')
                print('%10s %10s' % ('SM init', 'SM final'))
                print('%10f %10f' % (loc_SM, new_SM))
                self.forces_report()
                if self.stabavailable:
                    self.stabreport()
                print('===========================')
    def transp_byvec(self, vec): #transport pre-calculated forces by a 'vec' deallocation in CG
        if self.sld.runme:
            mom_transp=np.cross(vec, np.array([self.CX, self.CY, self.CZ]))
            self.Cl-=mom_transp[0]/self.bref
            self.Cm-=mom_transp[1]/self.cref
            self.Cn-=mom_transp[2]/self.bref
            if self.stabavailable:
                pars=['a', 'b', 'p', 'q', 'r']
                for par in pars:
                    dmom_transp=np.cross(vec, np.array([self.stabderivative_dict[par]['dCX'], self.stabderivative_dict[par]['dCY'], \
                        self.stabderivative_dict[par]['dCZ']]))
                    self.stabderivative_dict[par]['dCl']-=dmom_transp[0]/self.bref
                    self.stabderivative_dict[par]['dCm']-=dmom_transp[1]/self.cref
                    self.stabderivative_dict[par]['dCn']-=dmom_transp[2]/self.bref
            if self.hascorrections:
                mom_transp=np.cross(vec, np.array([self.dCX, self.dCY, self.dCZ]))
                self.dCl-=mom_transp[0]/self.bref
                self.dCm-=mom_transp[1]/self.cref
                self.dCn-=mom_transp[2]/self.bref
    def transp_to_cg(self, CG):
        if self.sld.runme:
            self.transp_byvec(-self.CG)
            self.transp_byvec(CG)
            self.CG=CG
    def __init__(self, sld, elems=[], Sref=0.0, cref=0.0, bref=0.0, echo=True, CG=np.array([0.0, 0.0, 0.0])):
        if sld.runme:
            self.wings=[]
            self.bodies=[]
            for e in elems:
                e.set_aircraft(self)
                if type(e)==wing:
                    self.wings+=[e]
                elif type(e)==body:
                    self.bodies+=[e]
            #BE VERY AWARE OF THESE WARNINGS WHEN USINGG FUSELAGE-SEPARATED SEMI-WINGS AS WING OBJECTS
            if len(self.wings)!=0:
                S, mac, b=self.wings[0].calc_reference()
            else:
                S=1.0; mac=1.0; b=1.0
            if Sref==0.0:
                print('WARNING: no reference surface provided. Using first wing input as reference. Address to aircraft constructor\'s Sref argument for better definitions')
                self.Sref=S
            else:
                self.Sref=Sref
            if cref==0.0:
                print('WARNING: no reference chord provided. Using first wing input as reference. Address to aircraft constructor\'s cref argument for better definitions')
                self.cref=mac
            else:
                self.cref=cref
            if bref==0.0:
                print('WARNING: no reference span provided. Using first wing input as reference. Address to aircraft constructor\'s bref argument for better definitions')
                self.bref=b
            else:
                self.bref=bref
            self.AR=(self.bref**2)/self.Sref

            self.a=0.0
            self.b=0.0
            self.p=0.0
            self.q=0.0
            self.r=0.0
            self.Uinf=1.0
            self.M=0.0
            self.rho=1.225
            self.mu=1.72e-5
            self.g=9.81
            self.gamma=1.4

            self.CG=CG
            
            #defining available controls
            self.controlset={}
            for wng in self.wings:
                for wngqd in wng.wingquads:
                    if wngqd.hascontrol():
                        for cname in wngqd.controls:
                            if not cname in self.controlset:
                                self.controlset[cname]=control_DOF()
                                wngqd.controls[cname].DOF=self.controlset[cname]

            self.sld=sld
            self.forcesavailable=False
            self.stabavailable=False
            self.massavailable=False
            self.haseqflatplate=False
            self.hascorrections=any([wng.hascorrection() for wng in self.wings])

            if echo:
                self.parameter_report()

            #defining plotting limits
            xkeypts=[]
            ykeypts=[]
            zkeypts=[]
            for wng in self.wings:
                for wngqd in wng.wingquads:
                    xkeypts+=[wngqd.sect1.CA_position[0]-wngqd.sect1.c/4, wngqd.sect1.CA_position[0]+3*wngqd.sect1.c/4, \
                        wngqd.sect2.CA_position[0]-wngqd.sect2.c/4, wngqd.sect2.CA_position[0]+3*wngqd.sect2.c/4]
                    ykeypts+=[wngqd.sect1.CA_position[1], wngqd.sect2.CA_position[1]]
                    zkeypts+=[wngqd.sect1.CA_position[2], wngqd.sect2.CA_position[2]]
            for bdy in self.bodies:
                xkeypts+=[bdy.sections[i].center[0] for i in range(len(bdy.sections))]
                ykeypts+=[bdy.sections[i].center[1]-bdy.sections[i].Rs[i] for i in range(len(bdy.sections))]
                zkeypts+=[bdy.sections[i].center[2]-bdy.sections[i].Rs[i] for i in range(len(bdy.sections))]
                ykeypts+=[bdy.sections[i].center[1]+bdy.sections[i].Rs[i] for i in range(len(bdy.sections))]
                zkeypts+=[bdy.sections[i].center[2]+bdy.sections[i].Rs[i] for i in range(len(bdy.sections))]
            if len(xkeypts)!=0:
                xmax=max(xkeypts)
                xmin=min(xkeypts)
                ymax=max(ykeypts)
                ymin=min(ykeypts)
                zmax=max(zkeypts)
                zmin=min(zkeypts)
                self.plotlim=max([abs(xmax), abs(ymax), abs(zmax), abs(xmin), abs(ymin), abs(zmin)])
            else:
                self.plotlim=1.0
        else:
            self.sld=sld
    def hascontrol(self):
        if self.sld.runme:
            return len(self.controlset)!=0
        else:
            return False
    def addwake(self, offset=1000.0, wakedisc=1, strategy=lambda x: ((np.exp(x)-1.0)/(exp(1)-1.0))**2):
        if self.sld.runme:
            #once the wake has been added, one can end geometry pre-processing of panels in question:
            self.sld.end_preprocess()
            for wng in self.wings:
                wng.genwakepanels(offset=offset, a=self.a, b=self.b, wakedisc=wakedisc, strategy=strategy)
    def calcforces(self, echo=True):
        if self.sld.runme:
            self.CX=sum([self.sld.forces[i][0] for i in range(self.sld.npanels)])/self.Sref
            self.CY=sum([self.sld.forces[i][1] for i in range(self.sld.npanels)])/self.Sref
            self.CZ=sum([self.sld.forces[i][2] for i in range(self.sld.npanels)])/self.Sref
            u=np.array([cos(self.a)*cos(self.b), sin(self.b)*cos(self.a), sin(self.a)])
            v=np.array([-sin(self.a)*cos(self.b), -sin(self.b)*sin(self.a), cos(self.a)])
            adforces=np.array([self.CX, self.CY, self.CZ])
            self.CL=v@adforces
            self.CD=u@adforces
            self.Cl=sum([self.sld.moments[i][0] for i in range(self.sld.npanels)])/(self.Sref*self.bref)
            self.Cm=sum([self.sld.moments[i][1] for i in range(self.sld.npanels)])/(self.Sref*self.cref)
            self.Cn=sum([self.sld.moments[i][2] for i in range(self.sld.npanels)])/(self.Sref*self.bref)

            if self.hascorrections:
                self.dCX=0.0; self.dCY=0.0; self.dCZ=0.0; self.dCL=0.0; self.dCD=0.0; self.dCl=0.0; self.dCm=0.0; self.dCn=0.0
                for wng in self.wings:
                    wng.calc_coefs(alpha=self.a, axis=wng.axis)
                    if wng.hascorrection():
                        dCX_wng, dCY_wng, dCZ_wng, dCl_wng, dCm_wng, dCn_wng=wng.calc_corrected_forces()
                        self.dCX+=dCX_wng
                        self.dCY+=dCY_wng
                        self.dCZ+=dCZ_wng
                        self.dCl+=dCl_wng
                        self.dCm+=dCm_wng
                        self.dCn+=dCn_wng
                adforces=np.array([self.dCX, self.dCY, self.dCZ])
                self.dCL=adforces@v
                self.dCD=adforces@u
            
            self.transp_byvec(self.CG)

            if echo:
                self.forces_report()
            self.forcesavailable=True
    def freestream_derivative(self, par='a'): #definition of derivative of freestream velocity vectors u and v (normal vectors)
        if par=='a':
            du=np.array([-sin(self.a)*cos(self.b), -sin(self.b)*sin(self.a), cos(self.a)])
            dv=np.array([-cos(self.a)*cos(self.b), -sin(self.b)*cos(self.a), -sin(self.a)])
        elif par=='b':
            du=np.array([-cos(self.a)*sin(self.b), cos(self.b)*cos(self.a), 0.0])
            dv=np.array([sin(self.a)*sin(self.b), -cos(self.b)*sin(self.a), 0.0])
        else:
            du=np.zeros(3)
            dv=np.zeros(3)
        return du, dv
    def calcstab(self, echo=True):
        if self.sld.runme:
            #calculate variation on Cps in respect to several parameters
            pars=['a', 'b', 'p', 'q', 'r']
            if self.forcesavailable:
                self.transp_byvec(-self.CG)
            self.stabderivative_dict={}
            orforces=np.array([self.CX, self.CY, self.CZ])
            for p in pars:
                dvdksi=self.sld.gen_farfield_derivative(self.Uinf, a=self.a, b=self.b, p=self.p*2*self.Uinf/self.bref, q=self.q*2*self.Uinf/self.cref, r=self.r*2*self.Uinf/self.bref, par=p)
                dCps=self.sld.calc_derivative_dv(self.Uinf, dvdksi)
                forces=[-dCps[i]*self.sld.panels[i].S*self.sld.panels[i].nvector for i in range(self.sld.npanels)]
                moments=[np.cross(self.sld.panels[i].colpoint, forces[i]) for i in range(self.sld.npanels)]
                dCX=sum([forces[i][0] for i in range(self.sld.npanels)])/self.Sref
                dCY=sum([forces[i][1] for i in range(self.sld.npanels)])/self.Sref
                dCZ=sum([forces[i][2] for i in range(self.sld.npanels)])/self.Sref
                u=np.array([cos(self.a)*cos(self.b), sin(self.b)*cos(self.a), sin(self.a)])
                v=np.array([-sin(self.a)*cos(self.b), -sin(self.b)*sin(self.a), cos(self.a)])
                adforces=np.array([dCX, dCY, dCZ])
                du, dv=self.freestream_derivative(par=p)
                dCL=v@adforces+orforces@dv
                dCD=u@adforces+orforces@du
                dCl=sum([moments[i][0] for i in range(self.sld.npanels)])/(self.Sref*self.bref)
                dCm=sum([moments[i][1] for i in range(self.sld.npanels)])/(self.Sref*self.cref)
                dCn=sum([moments[i][2] for i in range(self.sld.npanels)])/(self.Sref*self.bref)
                self.stabderivative_dict[p]={'dCX':dCX, 'dCY':dCY, 'dCZ':dCZ, 'dCL':dCL, 'dCD':dCD, 'dCl':dCl, 'dCm':dCm, 'dCn':dCn}
            self.stabavailable=True
            self.transp_byvec(self.CG)
            if echo:
                self.stabreport()
    def stabreport(self): #report calculated stability derivatives, or calculate them in place
        if self.sld.runme:
            if not self.stabavailable:
                self.calcstab(echo=False)
            print('Stability derivatives')
            print('%5s | %10s %10s %10s %10s %10s %10s %10s %10s' % ('par', 'dCX', 'dCY', 'dCZ', 'dCL', 'dCD', 'dCl', 'dCm', 'dCn'))
            for p in self.stabderivative_dict:
                print('%5s | %10f %10f %10f %10f %10f %10f %10f %10f' % (p, self.stabderivative_dict[p]['dCX'], self.stabderivative_dict[p]['dCY'], \
                    self.stabderivative_dict[p]['dCZ'], self.stabderivative_dict[p]['dCL'], self.stabderivative_dict[p]['dCD'], -self.stabderivative_dict[p]['dCl'], \
                        self.stabderivative_dict[p]['dCm'], self.stabderivative_dict[p]['dCn']))
    def edit_parameters(self, pardict, echo=True):
        if self.sld.runme:
            for par in pardict:
                val=pardict[par]
                if par=='a':
                    self.a=radians(val)
                elif par=='b':
                    self.b=radians(val)
                elif par=='p':
                    self.p=val
                elif par=='q':
                    self.q=val
                elif par=='r':
                    self.r=val
                elif par=='Uinf':
                    self.Uinf=val
                elif par=='M':
                    self.M=val
                elif par=='rho':
                    self.rho=val
                elif par=='mu':
                    self.mu=val
                elif par=='g':
                    self.g=val
                elif par=='gamma':
                    self.gamma=val
                else:
                    self.controlset[par].state=radians(val)
            #reset result readiness
            self.stabavailable=False
            self.forcesavailable=False
            self.haseqflatplate=False
            if echo:
                self.parameter_report()
    def bodies_eqflatplate_apply(self, rho=1.225, mu=1.72*10e-5, turbulent_criterion=Re2e5, Cf_l_rule=Blausius_Cf_l, Cf_t_rule=Prandtl_1_7th):
        if self.sld.runme:
            self.haseqflatplate=True
            for bdy in self.bodies:
                bdy.apply_eqflatplate(rho=rho, mu=mu, turbulent_criterion=turbulent_criterion, Cf_l_rule=Cf_l_rule, Cf_t_rule=Cf_t_rule, Uinf=self.Uinf)
    def eulersolve(self, echo=True, damper=0.0, tolerance=1e-5, wakeiter=0):
        if self.sld.runme:
            self.sld.eulersolve(target=np.array([]), a=self.a, b=self.b, p=self.p*2*self.Uinf/self.bref, q=self.q*2*self.Uinf/self.cref, \
                r=self.r*2*self.Uinf/self.bref, damper=damper, Uinf=self.Uinf, echo=echo, M=self.M, gamma=self.gamma, tolerance=tolerance, wakeiter=wakeiter, beta=sqrt(1.0-self.M**2))
    def resolve(self, echo=True, damper=0.0, tolerance=1e-5, wakeiter=0):
        if self.sld.runme:
            self.sld.resolve(target=np.array([]), a=self.a, b=self.b, p=self.p*2*self.Uinf/self.bref, q=self.q*2*self.Uinf/self.cref, \
                r=self.r*2*self.Uinf/self.bref, damper=damper, Uinf=self.Uinf, echo=echo, M=self.M, gamma=self.gamma, tolerance=tolerance, wakeiter=wakeiter, beta=sqrt(1.0-self.M**2))
    def forces_report(self):
        if self.sld.runme:
            print('========Total Forces Report========')
            if self.hascorrections:
                if self.haseqflatplate:
                    if self.haseqflatplate:
                        str_Cf='+local Cf estimation'
                else:
                    str_Cf=''
                print('Inviscid'+str_Cf+' case: ')
            if not self.forcesavailable:
                self.calcforces(echo=False)
            print('%10s %10s %10s %10s %10s %10s | %10s %10s' % ('CX', 'CY', 'CZ', 'Cl', 'Cm', 'Cn', 'CL', 'CD'))
            print('%10f %10f %10f %10f %10f %10f | %10f %10f' % (self.CX, self.CY, self.CZ, -self.Cl, self.Cm, -self.Cn, self.CL, self.CD))
            if self.hascorrections:
                print('After viscous external polar corrections: ')
                print('%10s %10s %10s %10s %10s %10s | %10s %10s' % ('CX', 'CY', 'CZ', 'Cl', 'Cm', 'Cn', 'CL', 'CD'))
                print('%10f %10f %10f %10f %10f %10f | %10f %10f' % (self.CX+self.dCX, self.CY+self.dCY, self.CZ+self.dCZ, \
                    -(self.Cl+self.dCl), self.Cm+self.dCm, -(self.Cn+self.dCn), self.CL+self.dCL, self.CD+self.dCD))
            print('===================================')
    def plot_input(self, xlim=[], ylim=[], zlim=[]):
        if self.sld.runme:
            fig=plt.figure()
            ax=plt.axes(projection='3d')
            for elem in self.bodies+self.wings:
                elem.plot_input(ax=ax, fig=fig)
            if len(xlim)!=0:
                ax.set_xlim3d(xlim[0], xlim[1])
            else:
                ax.set_xlim3d(-self.plotlim, self.plotlim)
            if len(ylim)!=0:
                ax.set_ylim3d(ylim[0], ylim[1])
            else:
                ax.set_ylim3d(-self.plotlim, self.plotlim)
            if len(zlim)!=0:
                ax.set_zlim3d(zlim[0], zlim[1])
            else:
                ax.set_zlim3d(-self.plotlim, self.plotlim)
            plt.xlabel('x')
            plt.ylabel('y')
            plt.show()
    def plotgeometry(self, xlim=[], ylim=[], zlim=[], factor=1.0, velfield=True):
        if self.sld.runme:
            if len(xlim)==0:
                xlim=[-self.plotlim, self.plotlim]
            if len(ylim)==0:
                ylim=[-self.plotlim, self.plotlim]
            if len(zlim)==0:
                zlim=[-self.plotlim, self.plotlim]
            self.sld.plotgeometry(xlim=xlim, ylim=ylim, zlim=zlim, factor=factor, velfield=velfield)
    def design_derivatives(self, wings=[], echo=True): #calculate derivatives in forces due to design parameters in a wing - its sectional incidences
        if self.sld.runme:
            if len(wings)==0:
                wings=self.wings
            wnglist=[]
            u=np.array([cos(self.a)*cos(self.b), sin(self.b)*cos(self.a), sin(self.a)])
            v=np.array([-sin(self.a)*cos(self.b), -sin(self.b)*sin(self.a), cos(self.a)])
            wngcnt=1
            for wng in wings:
                sectcnt=1
                wnglist+=[[]]
                if echo:
                    print('=============================================')
                    print('Design derivatives for incidences in wing %d [deg]' % (wngcnt))
                for sect in range(len(wng.wingquads)+1):
                    dFs, dMs=wng.calc_design_dFM(section=sect, Uinf=self.Uinf, M=self.M, gamma=self.gamma, rho=self.rho, CG=self.CG)
                    #transform to derivatives in respect to degrees
                    '''dFs*=np.pi/180.0
                    dMs*=np.pi/180.0'''
                    dFs/=self.Sref*self.rho*self.Uinf**2/2
                    dMs[0]/=self.Sref*self.bref*self.rho*self.Uinf**2/2
                    dMs[1]/=self.Sref*self.cref*self.rho*self.Uinf**2/2
                    dMs[2]/=self.Sref*self.bref*self.rho*self.Uinf**2/2
                    stwise=np.array([dFs@u, dFs@v])
                    derdict={'dCX':dFs[0], 'dCY':dFs[1], 'dCZ':dFs[2], 'dCD':stwise[0], 'dCL':stwise[1], 'dCl':-dMs[0], 'dCm':dMs[1], 'dCn':-dMs[2]} #stability derivatives
                    #transformed to derivative coordinate system
                    if echo:
                        print('===========Section %d==========' % (sectcnt))
                        print('%8s %8s %8s %8s %8s %8s %8s %8s' % ('dCX', 'dCY', 'dCZ', 'dCD', 'dCL', 'dCl', 'dCm', 'dCn'))
                        print('%8f %8f %8f %8f %8f %8f %8f %8f' % (derdict['dCX'], derdict['dCY'], derdict['dCZ'], derdict['dCD'], derdict['dCL'], \
                            derdict['dCl'], derdict['dCm'], derdict['dCn']))
                    wnglist[-1]+=[derdict]
                    sectcnt+=1
                if echo:
                    print('=============================================')
                wngcnt+=1
            return wnglist
        else:
            return []
    def dump(self, filename, filedir='', ext_append=True, erase_AIC=True, overwrite=True):
        if self.sld.runme:
            if len(filedir)==0:
                filedir=os.getcwd()
            fname=filename+('.lpm' if ext_append else '')
            ordir=os.getcwd()
            os.chdir(filedir)
            if os.path.exists(fname):
                if overwrite:
                    print('WARNING: aircraft file already exists. Overwritting')
                else:
                    raise Exception('Attempted to overwrite acft file with aircraft.dump() kwarg overwrite set to False')
            if erase_AIC:
                if hasattr(self.sld, 'aicm3') or hasattr(self.sld, 'aicm3_line') or hasattr(self.sld, 'aicm') or hasattr(self.sld, 'iaicm'):
                    print('WARNING: erasing AIC matrixes from solid for aircraft dumping. To override this behaviour, change kwarg flag erase_AIC to False.')
                    self.sld.aic_memory_clean()
            fil=open(fname, 'wb')
            pck.dump(self, fil)
            fil.close()
            os.chdir(ordir)

def read_aircraft(filename, filedir='', ext_append=True):
    if mp.current_process().name!='LPM_child': #guard against recursive imports
        if len(filedir)==0:
            filedir=os.getcwd()
        fname=filename+('.lpm' if ext_append else '')
        ordir=os.getcwd()
        os.chdir(filedir)
        if not os.path.exists(fname):
            raise Exception('File for aircraft loading not found: '+filedir+'/'+fname)
        fil=open(fname, 'rb')
        acft=pck.load(fil)
        fil.close()
        os.chdir(ordir)
        return acft
    else:
        sld=Solid() #this Solid, due to the current process's identity, should be returned as a runme==False Solid and lead to non-recursive imports
        return aircraft(sld, elems=[], echo=False)


import numpy as np
import numpy.linalg as lg
import scipy.linalg as slg
import scipy.linalg.lapack as slapack
from math import *
import matplotlib.pyplot as plt
from mpl_toolkits import mplot3d
import time as tm

import pytoolkit

from paneller import *
from utils import *
from multiprocess_guard import *

if multiprocess_guard():
    '''
    Script to test Euler solution with a NACA 4412 straight wing
    '''

    xdisc=20
    airfoil, extra, intra=read_afl('n4412', ext_append=True, remove_TE_gap=True, disc=xdisc, extra_intra=True)
    b=2.0
    c=0.3
    Uinf=0.2
    ys=np.linspace(b/2, -b/2, 20)
    #ys=np.sin(np.linspace(-pi/2, pi/2, 30))*b/2
    elip_factor=np.sqrt((b**2/4-0.9*ys**2)*4/b**2)
    #elip_factor=np.array([1.0]*len(ys))
    totlist=[]
    conlist=[]
    n=0
    for i in range(np.size(airfoil, 0)):
        totlist+=[[]]
        for j in range(len(ys)):
            totlist[-1]+=[np.array([airfoil[i, 0]*elip_factor[j]*c, ys[j], airfoil[i, 1]*elip_factor[j]*c])]
    for i in range(len(ys)-1):
        conlist+=[[i, i+(len(ys)-1)*(np.size(airfoil, 0)-2)]]
    sld=Solid(sldlist=[totlist], wraparounds=[[1]])
    sld.end_preprocess()
    sld.genwakepanels(wakecombs=conlist, a=radians(5.0))
    sld.plotgeometry(ylim=[-b/2, b/2], xlim=[-b/2, b/2], zlim=[-b/2, b/2])
    t=tm.time()
    sld.genvbar(Uinf, a=radians(5.0))
    sld.gennvv()
    print('Pre-processing and geometry gen: '+str(tm.time()-t))
    t=tm.time()
    sld.genaicm()
    print('AIC matrix generation: '+str(tm.time()-t))
    t=tm.time()
    sld.solve(damper=0.001)
    sld.calcpress(Uinf=Uinf)
    print('Solution and post-processing: '+str(tm.time()-t))
    sld.plotgeometry(ylim=[-b/2, b/2], xlim=[-b/2, b/2], zlim=[-b/2, b/2])
    #sld.plotgeometry()

    xlocs1=[]
    ylocs1=[]
    cps1=[]
    xlocs2=[]
    ylocs2=[]
    cps2=[]
    for i in range(sld.npanels):
        if sld.panels[i].nvector[2]>0:
            xlocs1+=[sld.panels[i].colpoint[0]]
            ylocs1+=[sld.panels[i].colpoint[1]]
            cps1+=[sld.Cps[i]]
        else:
            xlocs2+=[sld.panels[i].colpoint[0]]
            ylocs2+=[sld.panels[i].colpoint[1]]
            cps2+=[sld.Cps[i]]

    fig=plt.figure()
    ax=plt.axes(projection='3d')

    ax.scatter3D(xlocs1, ylocs1, cps1, 'red')
    ax.scatter3D(xlocs2, ylocs2, cps2, 'blue')
    plt.show()

    xlocs1=[]
    ylocs1=[]
    g1=[]
    xlocs2=[]
    ylocs2=[]
    g2=[]
    for i in range(sld.npanels):
        if sld.panels[i].nvector[2]>0:
            xlocs1+=[sld.panels[i].colpoint[0]]
            ylocs1+=[sld.panels[i].colpoint[1]]
            g1+=[sld.solution[i]]
        else:
            xlocs2+=[sld.panels[i].colpoint[0]]
            ylocs2+=[sld.panels[i].colpoint[1]]
            g2+=[sld.solution[i]]

    fig=plt.figure()
    ax=plt.axes(projection='3d')

    ax.scatter3D(xlocs1, ylocs1, g1, 'red')
    ax.scatter3D(xlocs2, ylocs2, g2, 'blue')
    plt.show()

    sld.plotnormals(ylim=[-b/2, b/2], xlim=[-b/2, b/2], zlim=[-b/2, b/2], factor=0.01)


import numpy as np
import numpy.linalg as lg
import scipy.linalg as slg
import scipy.linalg.lapack as slapack
from math import *
import matplotlib.pyplot as plt
from mpl_toolkits import mplot3d
import time as tm
import multiprocess as mp

from paneller import *
from utils import *
from wing import *
from body import *
from aircraft import *
from aerodynamic_output import *
from multiprocess_guard import *

if multiprocess_guard():
    Uinf=1.0
    t=tm.time()
    c=0.0
    thetas=np.linspace(0.0, 2*pi, 100)
    ys=np.linspace(-1000.0, 1000.0, 5)
    R=1.0
    coordlist1=[]

    for th in thetas:
        coordlist1+=[[]]
        for i in range(len(ys)):
            coordlist1[-1]+=[np.array([sin(th), ys[i], cos(th)])*R]
            '''
    for i in range(len(coordlist1)):
        coordlist1[i]=[np.array([0.0, ys[0], 0.0])]+coordlist1[i]+[np.array([0.0, ys[-1], 0.0])]'''
        

    sld=Solid()
    horzlines, vertlines, paninds, sldpts=sld.addpatch(coordlist1, wraps=[1])

    sld.end_preprocess()

    sld.genvbar(Uinf=Uinf)
    sld.gennvv()
    print('Solid generation and pre-processing: '+str(tm.time()-t))
    t=tm.time()
    sld.genaicm()
    sld.gen_selfinf_mat()
    print('Generating AICs: '+str(tm.time()-t))
    t=tm.time()
    sld.solve(damper=c)
    print('Solving and post-processing: '+str(tm.time()-t))

    sld.plotgeometry()
    plt.scatter([p.colpoint[0] for p in sld.panels], \
        [lg.norm(sld.delphi[i, :]+sld.vbar[i, :])/Uinf \
        for i in range(len(sld.panels))])
    plt.xlabel('$x$')
    plt.ylabel('$v$')
    plt.show()



import multiprocessing as mp

def multiprocess_guard(): #return False if code is being executed from LPM's child processes
    return mp.current_process().name!='LPM_child' #built to be implemented as a more proper if __name__=='__main__' guard for Windows multiprocessing


import numpy as np
import numpy.linalg as lg
import scipy.linalg as slg
import scipy.linalg.lapack as slapack
from math import *
import matplotlib.pyplot as plt
from mpl_toolkits import mplot3d
import time as tm

import pytoolkit

from paneller import *
from multiprocess_guard import *

if multiprocess_guard():
    '''
    Script made to test the Euler solution modules with a flat-plate straight wing
    '''

    b=1.0
    c=1.0

    u=0.05

    ys=np.linspace(-b/2, b/2, 50)
    xs=np.linspace(0.0, c, 50)

    wakeon=[]
    coords=[]
    npan=0
    geomlist=[]
    for i in range(len(xs)-1):
        coords=[]
        for j in range(len(ys)-1):
            coords+=[np.array([[xs[i+1], xs[i+1], xs[i], xs[i]], \
                [ys[j], ys[j+1], ys[j+1], ys[j]], [0.0, 0.0, 0.0, 0.0]])]
            if i==len(xs)-2:
                wakeon+=[[npan, -1]]
            npan+=1
        geomlist+=[coords]
    sld=Solid(sldlist=[geomlist])
    sld.genwakepanels(wakecombs=wakeon, a=radians(30.0))
    sld.genvbar(u, a=radians(30.0))
    sld.gennvv()
    t=tm.time()
    sld.genaicm()
    print(tm.time()-t)
    t=tm.time()
    sld.solve(damper=0.0)
    print(tm.time()-t)
    sld.calcpress(Uinf=u)
    sld.plotpress()

    sld.plotgeometry(wake=False)

    sld.calcforces()
    print('CFres: '+str(sld.SCFres))
    print('CMres: '+str(sld.SCMres))

    fig=plt.figure()
    ax=plt.axes(projection='3d')

    xs=[]
    ys=[]
    for i in range(sld.npanels):
        xs+=[sld.panels[i].colpoint[0]]
        ys+=[sld.panels[i].colpoint[1]]
    ax.scatter3D(xs, ys, sld.Cps)
    plt.show()


import os
import sys
import datetime
import pickle
import LoveUpdate
ordir=os.getcwd()
os.chdir(os.path.dirname(os.path.abspath(__file__)))
sys.path.insert(0, os.getcwd())
from multiprocess_guard import *
from utils import *
from wing import *
from body import *
from control import *
from aircraft import *
from xfoil_visc import *
from aerodynamic_output import *
from pytoolkit import *

#Define update module control functions
def updateset():
    pdir=os.getcwd()
    os.chdir(os.path.dirname(os.path.abspath(__file__)))
    lupfile=open('updatestat.lup', 'rb')
    statusdict=pickle.load(lupfile)
    lupfile.close()
    statusdict['updateme']=True
    lupfile=open('updatestat.lup', 'wb')
    pickle.dump(statusdict, lupfile)
    lupfile.close()
    os.chdir(pdir)
def updatecancel():
    pdir=os.getcwd()
    os.chdir(os.path.dirname(os.path.abspath(__file__)))
    lupfile=open('updatestat.lup', 'rb')
    statusdict=pickle.load(lupfile)
    lupfile.close()
    statusdict['updateme']=False
    lupfile=open('updatestat.lup', 'wb')
    pickle.dump(statusdict, lupfile)
    lupfile.close()
    os.chdir(pdir)
def update_message_enable():
    pdir=os.getcwd()
    os.chdir(os.path.dirname(os.path.abspath(__file__)))
    lupfile=open('updatestat.lup', 'rb')
    statusdict=pickle.load(lupfile)
    lupfile.close()
    statusdict['message']=True
    lupfile=open('updatestat.lup', 'wb')
    pickle.dump(statusdict, lupfile)
    lupfile.close()
    os.chdir(pdir)
def update_message_suppress():
    pdir=os.getcwd()
    os.chdir(os.path.dirname(os.path.abspath(__file__)))
    lupfile=open('updatestat.lup', 'rb')
    statusdict=pickle.load(lupfile)
    lupfile.close()
    statusdict['message']=False
    lupfile=open('updatestat.lup', 'wb')
    pickle.dump(statusdict, lupfile)
    lupfile.close()
    os.chdir(pdir)

if multiprocess_guard():
    #Generate update registry file if none is found
    tdy=LoveUpdate.getdate()
    if not os.path.exists('updatestat.lup'):
        print('NOTICE: update registry file not found. Creating anew')
        lupfile=open('updatestat.lup', 'wb')
        statusdict={'lastupdate':tdy, 'updateme':True, 'message':True}
        _=LoveUpdate.update('LoveUpdate', '')
        pickle.dump(statusdict, lupfile)
        lupfile.close()

    #Read update registry file
    lupfile=open('updatestat.lup', 'rb')
    statusdict=pickle.load(lupfile)
    lupfile.close()

    #Declare variables
    hasupdated=False
    hasupdated_lup=False
    code_lup=False
    code=False

    #If an older version update file doesn't include message suppression/enabling variable, create it
    if not 'message' in statusdict:
        update_message_enable() #activate message issuing when autoupdate is not performed
        statusdict['message']=True

    #Update if activated
    if statusdict['updateme']:
        code_lup, hasupdated_lup, tdy=LoveUpdate.update('LoveUpdate', statusdict['lastupdate'])
        if not code_lup:
            if hasupdated_lup:
                print('LoveUpdate has autoupdated. If you wish to deactivate the autoupdate feature, import LovelacePM in another python instance and use LovelacePM.updatecancel()')
            code, hasupdated, tdy=LoveUpdate.update('LovelacePM', statusdict['lastupdate'])
            code=code and code_lup
            if not code:
                if hasupdated:
                    print('LovelacePM has autoupdated. If you wish to deactivate the autoupdate feature, import LovelacePM in another python instance and use LovelacePM.updatecancel()')
                statusdict['lastupdate']=tdy
                lupfile=open('updatestat.lup', 'wb')
                pickle.dump(statusdict, lupfile)
                lupfile.close()
            else:
                print('WARNING: error in autoupdate. Check internet connection if you wish to update LovelacePM. Moving on')
        else:
            print('WARNING: error in autoupdate. Check internet connection if you wish to update LoveUpdate. Moving on')
    elif statusdict['message']: #warn if message flag is activated
        print('WARNING: LovelacePM autoupdate is deactivated. if you wish to activate autoupdate, import LovelacePM in another python instance and use LovelacePM.updateset()')
        print('To suppress this message, use LovelacePM.update_message_suppress(). To undo this action, use LovelacePM.update_message_enable()')


    del code, code_lup, lupfile, tdy, hasupdated_lup, hasupdated
os.chdir(ordir)
del ordir



from LovelacePM import *
import numpy as np

if multiprocess_guard():
    ordir=os.getcwd()
    os.chdir(os.path.dirname(os.path.abspath(__file__)))

    c=0.5
    AR=20.0
    b=c*AR
    nu=1.72e-5
    rho=1.225
    Uinf=10.0
    Re=Uinf*rho*c

    a=0.0

    n0012_visc=polar_correction('n0012', aseq=[-5.0, 5.0, 0.5])

    sld=Solid()

    sl=wing_section(afl='n0012', c=c, CA_position=np.array([0.0, -b/2, 0.0]), xdisc=30, closed=True, correction=n0012_visc, Re=Re)
    sr=wing_section(afl='n0012', c=c, CA_position=np.array([0.0, b/2, 0.0]), xdisc=30, closed=True, correction=n0012_visc, Re=Re)
    wngqd=wing_quadrant(sld, sect1=sl, sect2=sr)
    wng=wing(sld, wingquads=[wngqd])
    acft=aircraft(sld, elems=[wng])

    acft.edit_parameters({'a':a, 'Uinf':Uinf})

    wng.patchcompose(ydisc=30)
    acft.addwake()

    acft.eulersolve()
    acft.forces_report()

    plot_Cps(sld, elems=[wng])

    os.chdir(ordir)


import numpy as np
import numpy.linalg as lg
import scipy.linalg as slg
import scipy.linalg.lapack as slapack
from math import *
import matplotlib.pyplot as plt
from mpl_toolkits import mplot3d
import time as tm
import numpy.random as rnd

import pytoolkit

from paneller import *
from multiprocess_guard import *

if multiprocess_guard():
    pan1=rnd.random(12).reshape(3, 4)
    pan2=rnd.random(12).reshape(3, 4)

    '''pan1=np.array([[0.59116381, 0.06286042, 0.71395661, 0.94342679], \
        [0.20698101, 0.25343613, 0.65463535, 0.9292159 ], \
            [0.99920911, 0.43311108, 0.57953426, 0.86468112]])
    pan2=np.array([[0.02830673, 0.45917184, 0.40259338, 0.39839244], \
    [0.67288593, 0.82413215, 0.31487439, 0.73770071], \
    [0.0768672,  0.13285883, 0.13590843, 0.81177605]])'''
    print(pan1)
    print(pan2)

    sld=Solid(sldlist=[pan1, pan2])
    sld.genaicm()
    print(sld.aicm3)


from LovelacePM import *
import numpy as np

if multiprocess_guard()
    ordir=os.getcwd()
    os.chdir(os.path.dirname(os.path.abspath(__file__)))

    #test case to compare with PMARC's test case of a rotating AR 5 NACA 0012 wing

    c=1.0
    AR=5.0
    b=c*AR
    Uinf=0.3048*1.0 #

    a=0.0
    p=radians(3)*Uinf/(b*2)

    sld=Solid()

    sl=wing_section(afl='n0012', c=c, CA_position=np.array([0.0, -b/2, 0.0]), xdisc=50, closed=True)
    sr=wing_section(afl='n0012', c=c, CA_position=np.array([0.0, b/2, 0.0]), xdisc=50, closed=True)
    wngqd=wing_quadrant(sld, sect1=sl, sect2=sr)
    wng=wing(sld, wingquads=[wngqd])
    acft=aircraft(sld, elems=[wng])

    acft.edit_parameters({'a':a, 'Uinf':Uinf, 'p':p})

    wng.patchcompose(ydisc=30)
    acft.addwake(wakedisc=50, strategy=lambda x: x, offset=b*10)

    acft.eulersolve(wakeiter=1)

    plot_Cps(sld, elems=[wng], zlim=[1.0, -1.0])

    acft.plotgeometry(velfield=False)

    os.chdir(ordir)


import numpy as np
import numpy.linalg as lg
import scipy.linalg as slg
import scipy.linalg.lapack as slapack
from math import *
import matplotlib.pyplot as plt
from mpl_toolkits import mplot3d
import time as tm
import multiprocess as mp

from paneller import *
from multiprocess_guard import *

'''
Script to test the Euler solution modules with a generic sphere geometry
'''
if multiprocess_guard():
    Uinf=1.0
    t=tm.time()
    c=0.0001
    thetas=np.linspace(0.0, 2*pi, 20)
    phis=np.linspace(-pi/2, pi/2, 100)
    R=1.0
    coordlist=[]
    conlist=[]
    n=0
    thlast=0
    exc=1.0
    for th in thetas:
        coordlist+=[[]]
        for phi in phis:
            coordlist[-1]+=[np.array([sin(phi)*exc, cos(phi)*cos(th), cos(phi)*sin(th)])*R]
    for i in range(len(phis)-1):
        for j in range(len(thetas)-1):
            if j==0:
                thlast=n
            if j==len(thetas)-2:
                conlist+=[[thlast, n]]
            n+=1
    sld=Solid(sldlist=[coordlist], wraparounds=[[1]])
    sld.end_preprocess()
    #sld.genwakepanels(wakecombs=conlist)
    sld.genvbar(Uinf=Uinf)
    sld.gennvv()
    print('Solid generation and pre-processing: '+str(tm.time()-t))
    t=tm.time()
    sld.genaicm()
    sld.gen_selfinf_mat()

    #checking aicm with srivastava's paper
    '''print(sld.aicm3_line)
    srivastava_order=np.array([3, 2, 0, 1])
    equivalences=np.array([6, 7, 4, 5])
    ptinds=np.arange(0, 4, dtype='int')
    aicm=(sld.aicm[0:4, equivalences]+sld.aicm[0:4, ptinds])
    for i in range(sld.npanels):
        print(i)
        lines, dumpme=np.nonzero(sld.panline_matrix[:, i])
        for l in lines:
            print(sld.lines[l, :, :])
        print(sld.panels[i].colpoint)
    aicm=aicm[srivastava_order, :]
    aicm=aicm[:, srivastava_order]
    print(aicm)'''

    print('Generating AICs: '+str(tm.time()-t))
    t=tm.time()

    sld.solve(damper=c)
    print('Solving and post-processing: '+str(tm.time()-t))

    '''for l in range(sld.nlines):
        for p in range(sld.npanels):
            print('Line')
            print(sld.lines[l, :, :])
            print('to panel')
            print(sld.panels[p].colpoint)
            print('n')
            print(sld.panels[p].nvector)
            print(sld.aicm3_line[:, p, l])
            print(sld.panels[p].nvector@sld.aicm3_line[:, p, l])'''

    sld.plotgeometry()
    plt.scatter([p.colpoint[0] for p in sld.panels], \
        [lg.norm(sld.delphi[i, :]+sld.vbar[i, :])/Uinf \
        for i in range(len(sld.panels))])
    plt.xlabel('$x$')
    plt.ylabel('$v$')
    plt.show()


import numpy as np
from math import *
import numpy.linalg as lg

#module to replace fdyn.f90

#solve aircraft positions as they vary in time
def tstep_solve(nstep, dt, rho, U0, g, Sref, cref, bref, perturbation, onboard, inertia, m, \
    coeffs, derivs):
    external_history=np.zeros((nstep, 3))
    euler_history=np.zeros((nstep, 3))
    alpha_history=np.zeros(nstep)
    beta_history=np.zeros(nstep)
    time_history=np.linspace(0.0, dt*nstep, nstep)
    a=0.0
    b=0.0
    Uinf=0.0
    forces=np.zeros(3)
    moments=np.zeros(3)
    mtemp=np.zeros(3)
    #for pertutbation, here we use the notation employed by Drela in Flight Viechle Aerodynamics
    #equations are also directly obtained from there
    #notation for coefs: (CX, CY, CZ, Cl, Cm, Cn)
    #notation for derivs: (CX, CY, CZ, Cl, Cm, Cn), first row: a derivatives; second, b derivatives; third, p; 4, q; 5, r
    xe=perturbation[0]
    ye=perturbation[1]
    ze=perturbation[2]
    phi=perturbation[3]
    theta=perturbation[4]
    psi=perturbation[5]
    u=perturbation[6]+U0
    v=perturbation[7]
    w=perturbation[8]
    p=perturbation[9]*2*U0/bref
    q=perturbation[10]*2*U0/cref
    r=perturbation[11]*2*U0/bref #denormalizing angular velocities
    invI=lg.inv(inertia)
    for i in range(nstep):
        Uinf=sqrt(u**2+v**2+w**2)
        a=atan(w/u)
        b=atan(v/sqrt(u**2+w**2))
        for j in range(3):
            forces[j]=(coeffs[j]+derivs[0, j]*a+derivs[1, j]*b+derivs[2, j]*p*bref/(2*Uinf)+\
                derivs[3, j]*q*cref/(2*Uinf)+derivs[4, j]*r*bref/(2*Uinf))*Sref*Uinf**2*rho/2
            moments[j]=(coeffs[j+3]+derivs[0, j+3]*a+derivs[1, j+3]*b+derivs[2, j+3]*p*bref/(2*Uinf)+\
                derivs[3, j+3]*q*cref/(2*Uinf)+derivs[4, j+3]*r*bref/(2*Uinf))*Sref*(cref if j==1 else bref)*Uinf**2*rho/2
        u=u+dt*(forces[0]/m-g*sin(theta)-q*w+r*v)
        v=v+dt*(forces[1]/m+g*sin(phi)*cos(theta)-r*u+p*w)
        w=w+dt*(forces[2]/m+g*cos(phi)*cos(theta)-p*v+q*u)
        mtemp[0]=moments[0]-(inertia[2, 2]-inertia[1, 1])*q*r-inertia[1, 2]*(q**2-r**2)-\
            inertia[0, 2]*p*q+inertia[0, 1]*p*r-onboard[2]*q+onboard[1]*r
        mtemp[1]=moments[1]-(inertia[0, 0]-inertia[2, 2])*r*p-inertia[0, 2]*(r**2-p**2)-\
            inertia[0, 1]*q*r+inertia[1, 2]*q*p-onboard[0]*r+onboard[2]*p
        mtemp[2]=moments[2]-(inertia[2, 2]-inertia[0, 0])*p*q-inertia[0, 1]*(p**2-q**2)-\
            inertia[1, 2]*r*p+inertia[0, 2]*r*q-onboard[1]*p+onboard[0]*q
        mtemp=invI@mtemp
        p+=dt*mtemp[0]
        q+=dt*mtemp[1]
        r+=dt*mtemp[2]
        phi=phi+dt*(p+q*sin(phi)*tan(theta)+r*cos(phi)*tan(theta))
        theta=theta+dt*(q*cos(phi)-r*sin(phi))
        psi=psi+dt*(q*sin(phi)/cos(theta)+r*cos(phi)/cos(theta))
        alpha_history[i]=a
        beta_history[i]=b
        euler_history[i, 0]=phi
        euler_history[i, 1]=theta
        euler_history[i, 2]=psi
        external_history[i, 0]=xe
        external_history[i, 1]=ye
        external_history[i, 2]=ze
    return external_history, alpha_history, beta_history, euler_history, time_history